# Copyright (c) 2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging
from oslo_serialization import jsonutils

from nova.scheduler import filters


LOG = logging.getLogger(__name__)


class CpuFlagsFilter(filters.BaseHostFilter):

    def host_passes(self, host_state, spec_obj):
        """Return a list of hosts that are compatible by cpu.

        This filter does not take into account instance cpu info.
        For example with cpu_mode = custom, instance cpu features
        may be compatible with the target host but this filter
        would reject a candidate due to source and target host
        incompatibility.
        """

        if ('scheduler_hints' not in spec_obj
            or not spec_obj.scheduler_hints.get("__operation")):
            return True
        operation = spec_obj.scheduler_hints.get("__operation")
        if operation[0] != "live-migration":
            return True
        LOG.debug("Applying CpuFlagsFilter on host (%s)", host_state.host)
        src_cpu_info = jsonutils.loads(operation[1])
        src_features = set(src_cpu_info['features'])
        cpu_info = jsonutils.loads(host_state.cpu_info)
        features = set(cpu_info['features'])
        if src_cpu_info['arch'] != cpu_info['arch']:
            LOG.debug("Host (%s) is incompatible by cpu arch", host_state.host)
            return False
        if not src_features.issubset(features):
            LOG.debug("Host (%s) is incompatible by cpu features",
                      host_state.host)
            return False
        return True
